package dashboard2014;

import edu.wpi.first.wpilibj.networktables2.NetworkTableNode;
import edu.wpi.first.wpilibj.networktables2.client.NetworkTableClient;
import edu.wpi.first.wpilibj.networktables2.stream.SocketStreamFactory;
import edu.wpi.first.wpilibj.tables.IRemote;
import edu.wpi.first.wpilibj.tables.IRemoteConnectionListener;
import edu.wpi.first.wpilibj.tables.ITable;
import edu.wpi.first.wpilibj.tables.ITableListener;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.SwingUtilities;
import javax.swing.Timer;

/**
 *
 * @author Robotics
 */
public class PingTask implements ActionListener, IRemoteConnectionListener, ITableListener {

    public static final boolean LOUD = false;
    public static final int CHECK_PERIOD = 500;
    public static final int PING_TIMEOUT = 300;
    public static final int CONNECTION_TIMEOUT = 1000;

    private volatile int ip;
    private volatile boolean ipChange;

    private final Backend backend;
    private boolean connecting;
    private NetworkTableClient client;

    public PingTask(int startIP, Backend b) {
        ip = startIP;
        backend = b;
        connecting = false;
        client = null;
        Timer t = new Timer(CHECK_PERIOD, this);
        t.setCoalesce(true);
        t.start();
    }

    public static boolean portPing(String ip, int timeout, int port) {
        // need a way to test this on robots;
        if (0 == 0) { // TODO: is this ping even necessary?? We can just try/fail a connection each time, like the camera
            return true;
        }
        try {
            return InetAddress.getByName(ip).isReachable(timeout);
        } catch (UnknownHostException ex) {
            Logger.getLogger(Backend.class.getName()).log(Level.SEVERE, null, ex);
        } catch (IOException ex) {
            Logger.getLogger(Backend.class.getName()).log(Level.SEVERE, null, ex);
        }
        return false;
    }

    @Override
    public void actionPerformed(ActionEvent e) {
        boolean available = portPing(Backend.getRobotIP(ip), 1735, PING_TIMEOUT);
        log(String.format("Ping success: %c\n", available ? '✓' : 'X'));
        if (ipChange) {
            ipChange = false;
            if (client != null) {
                disconnect();
            }
            if (available) {
                connect();
            }
        } else {
            if (available) {
                if (client == null && !connecting) {
                    connect();
                }
            } else {
                if (client != null) {
                    disconnect();
                }
            }
        }
    }

    public void changeIP(int newIP) {
        if (newIP == ip) {
            return;
        }
        ip = newIP;
        ipChange = true;
    }

    public void disconnect() {
        final NetworkTableClient cl = client;
        client = null;

        cl.removeConnectionListener(this);
        cl.removeTableListener(this);
        disconnected(null);
        (new Thread(new Runnable() {
            @Override
            public void run() {
                log("Disconnecting client");
                cl.stop();
                log("Done disconnecting client");
            }
        })).start();
    }

    public static void waitUntilConnected(final NetworkTableClient client, long timeout) {
        final AtomicBoolean lock = new AtomicBoolean(false);
        IRemoteConnectionListener listener = new IRemoteConnectionListener() {
            @Override
            public void connected(IRemote remote) {
                lock.getAndSet(true);
            }

            @Override
            public void disconnected(IRemote remote) {
                lock.getAndSet(false);
            }
        };
        client.addConnectionListener(listener, true);
        client.reconnect();
        long start = System.currentTimeMillis();
        while (!lock.get() && System.currentTimeMillis() < start + CONNECTION_TIMEOUT) {
            try {
                Thread.sleep(50);
            } catch (InterruptedException ex) {
                Logger.getLogger(PingTask.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        client.removeConnectionListener(listener);
    }

    public void connect() {
        connecting = true;
        final int connIP = ip;
        final PingTask self = this;
        (new Thread(new Runnable() {
            @Override
            public void run() {
                log("Creating a client");
                NetworkTableClient cl;
                try {
                    cl = new NetworkTableClient(new SocketStreamFactory(Backend.getRobotIP(connIP), 1735));
                } catch (IOException e) {
                    log("Making a client failed.");
                    connecting = false;
                    return;
                }

                waitUntilConnected(cl, 1000);

                if (cl.isConnected()) {
                    if (ip == connIP) {
                        log("IP still relevant.");
                        client = cl;
                        connecting = false;
                        client.addConnectionListener(self, true);
                        client.addTableListener(self, false);
                        log("Connection successful.");
                        return;
                    }
                    log("IP change: aborting.");
                } else {
                    log("Connection failed: aborting.");
                }

                // ip is no longer valid; we abort
                connecting = false;
                cl.stop();
                log("Done aborting connection.");

            }
        })).start();
    }

    /**
     * @return Maybe null
     */
    public NetworkTableNode getTable() {
        return client;
    }

    @Override
    public void connected(IRemote remote) {
        final NetworkTableClient cl = client;
        SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                backend.connected(cl);
            }
        });
    }

    @Override
    public void disconnected(IRemote remote) {
        SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                backend.disconnected();
            }
        });
    }

    @Override
    public void valueChanged(final ITable source, final String key, final Object value, boolean isNew) {
        final NetworkTableClient cl = client;
        SwingUtilities.invokeLater(new Runnable() {
            @Override
            public void run() {
                backend.valueChanged(cl, key, value);
            }
        });
    }

    private void log(String x) {
        if (LOUD) {
            System.err.println(x);
        }
    }
}
